@@ -6,7 +6,9 @@
 
 #include "dm.h"
 
+#include <linux/module.h>
 #include <linux/kmod.h>
+#include <linux/slab.h>
 
 struct tt_internal {
 	struct target_type tt;
@@ -16,61 +18,10 @@ struct tt_internal {
 };
 
 static LIST_HEAD(_targets);
-static rwlock_t _lock = RW_LOCK_UNLOCKED;
+static DECLARE_RWSEM(_lock);
 
 #define DM_MOD_NAME_SIZE 32
 
-/*
- * Destructively splits up the argument list to pass to ctr.
- */
-int split_args(int max, int *argc, char **argv, char *input)
-{
-	char *start, *end = input, *out;
-	*argc = 0;
-
-	while (1) {
-		start = end;
-
-		/* Skip whitespace */
-		while (*start && isspace(*start))
-			start++;
-
-		if (!*start)
-			break;	/* success, we hit the end */
-
-		/* 'out' is used to remove any back-quotes */
-		end = out = start;
-		while (*end) {
-			/* Everything apart from '\0' can be quoted */
-			if (*end == '\\' && *(end + 1)) {
-				*out++ = *(end + 1);
-				end += 2;
-				continue;
-			}
-
-			if (isspace(*end))
-				break;	/* end of token */
-
-			*out++ = *end++;
-		}
-
-		/* have we already filled the array ? */
-		if ((*argc + 1) > max)
-			return -EINVAL;
-
-		/* we know this is whitespace */
-		if (*end)
-			end++;
-
-		/* terminate the string and put it in the array */
-		*out = '\0';
-		argv[*argc] = start;
-		(*argc)++;
-	}
-
-	return 0;
-}
-
 static inline struct tt_internal *__find_target_type(const char *name)
 {
 	struct list_head *tih;
@@ -90,7 +41,7 @@ static struct tt_internal *get_target_type(const char *name)
 {
 	struct tt_internal *ti;
 
-	read_lock(&_lock);
+	down_read(&_lock);
 	ti = __find_target_type(name);
 
 	if (ti) {
@@ -98,7 +49,7 @@ static struct tt_internal *get_target_type(const char *name)
 			__MOD_INC_USE_COUNT(ti->tt.module);
 		ti->use++;
 	}
-	read_unlock(&_lock);
+	up_read(&_lock);
 
 	return ti;
 }
@@ -113,8 +64,6 @@ static void load_module(const char *name)
 
 	strcat(module_name, name);
 	request_module(module_name);
-
-	return;
 }
 
 struct target_type *dm_get_target_type(const char *name)
@@ -133,13 +82,13 @@ void dm_put_target_type(struct target_type *t)
 {
 	struct tt_internal *ti = (struct tt_internal *) t;
 
-	read_lock(&_lock);
+	down_read(&_lock);
 	if (--ti->use == 0 && ti->tt.module)
 		__MOD_DEC_USE_COUNT(ti->tt.module);
 
 	if (ti->use < 0)
 		BUG();
-	read_unlock(&_lock);
+	up_read(&_lock);
 
 	return;
 }
@@ -156,6 +105,20 @@ static struct tt_internal *alloc_target(struct target_type *t)
 	return ti;
 }
 
+
+int dm_target_iterate(void (*iter_func)(struct target_type *tt,
+					void *param), void *param)
+{
+	struct tt_internal *ti;
+
+	down_read(&_lock);
+	list_for_each_entry (ti, &_targets, list)
+		iter_func(&ti->tt, param);
+	up_read(&_lock);
+
+	return 0;
+}
+
 int dm_register_target(struct target_type *t)
 {
 	int rv = 0;
@@ -164,13 +127,14 @@ int dm_register_target(struct target_type *t)
 	if (!ti)
 		return -ENOMEM;
 
-	write_lock(&_lock);
-	if (__find_target_type(t->name))
+	down_write(&_lock);
+	if (__find_target_type(t->name)) {
+		kfree(ti);
 		rv = -EEXIST;
-	else
+	} else
 		list_add(&ti->list, &_targets);
 
-	write_unlock(&_lock);
+	up_write(&_lock);
 	return rv;
 }
 
@@ -178,53 +142,50 @@ int dm_unregister_target(struct target_type *t)
 {
 	struct tt_internal *ti;
 
-	write_lock(&_lock);
+	down_write(&_lock);
 	if (!(ti = __find_target_type(t->name))) {
-		write_unlock(&_lock);
+		up_write(&_lock);
 		return -EINVAL;
 	}
 
 	if (ti->use) {
-		write_unlock(&_lock);
+		up_write(&_lock);
 		return -ETXTBSY;
 	}
 
 	list_del(&ti->list);
 	kfree(ti);
 
-	write_unlock(&_lock);
+	up_write(&_lock);
 	return 0;
 }
 
 /*
  * io-err: always fails an io, useful for bringing
- * up LV's that have holes in them.
+ * up LVs that have holes in them.
  */
-static int io_err_ctr(struct dm_table *t, offset_t b, offset_t l,
-		      int argc, char **args, void **context)
+static int io_err_ctr(struct dm_target *ti, unsigned int argc, char **args)
 {
-	*context = NULL;
 	return 0;
 }
 
-static void io_err_dtr(struct dm_table *t, void *c)
+static void io_err_dtr(struct dm_target *ti)
 {
 	/* empty */
-	return;
 }
 
-static int io_err_map(struct buffer_head *bh, int rw, void *context)
+static int io_err_map(struct dm_target *ti, struct buffer_head *bh, int rw,
+		      union map_info *map_context)
 {
-	buffer_IO_error(bh);
-	return 0;
+	return -EIO;
 }
 
 static struct target_type error_target = {
-	name:	"error",
-	ctr:	io_err_ctr,
-	dtr:	io_err_dtr,
-	map:	io_err_map,
-	status:	NULL,
+	.name = "error",
+	.version = {1, 0, 1},
+	.ctr  = io_err_ctr,
+	.dtr  = io_err_dtr,
+	.map  = io_err_map,
 };
 
 int dm_target_init(void)
