@@ -56,10 +56,13 @@
 #define PROC_DEVICES "/proc/devices"
 #define MISC_NAME "misc"
 
+#define NUMBER_OF_MAJORS 4096
+
 /* dm major version no for running kernel */
 static int _dm_version = DM_VERSION_MAJOR;
 static int _log_suppress = 0;
 
+static dm_bitset_t _dm_bitset = NULL;
 static int _control_fd = -1;
 static int _version_checked = 0;
 static int _version_ok = 1;
@@ -119,12 +122,17 @@ static void *_align(void *ptr, unsigned int a)
 }
 
 #ifdef DM_IOCTLS
+/*
+ * Set number to NULL to populate _dm_bitset - otherwise first
+ * match is returned.
+ */
 static int _get_proc_number(const char *file, const char *name,
 			    uint32_t *number)
 {
 	FILE *fl;
 	char nm[256];
 	int c;
+	uint32_t num;
 
 	if (!(fl = fopen(file, "r"))) {
 		log_error("%s: fopen failed: %s", file, strerror(errno));
@@ -132,10 +140,14 @@ static int _get_proc_number(const char *file, const char *name,
 	}
 
 	while (!feof(fl)) {
-		if (fscanf(fl, "%d %255s\n", number, &nm[0]) == 2) {
+		if (fscanf(fl, "%d %255s\n", &num, &nm[0]) == 2) {
 			if (!strcmp(name, nm)) {
-				fclose(fl);
-				return 1;
+				if (number) {
+					*number = num;
+					fclose(fl);
+					return 1;
+				}
+				dm_bit_set(_dm_bitset, num);
 			}
 		} else do {
 			c = fgetc(fl);
@@ -143,8 +155,12 @@ static int _get_proc_number(const char *file, const char *name,
 	}
 	fclose(fl);
 
-	log_error("%s: No entry for %s found", file, name);
-	return 0;
+	if (number) {
+		log_error("%s: No entry for %s found", file, name);
+		return 0;
+	}
+
+	return 1;
 }
 
 static int _control_device_number(uint32_t *major, uint32_t *minor)
@@ -230,6 +246,35 @@ static int _create_control(const char *control, uint32_t major, uint32_t minor)
 }
 #endif
 
+static int _create_dm_bitset(void)
+{
+#ifdef DM_IOCTLS
+	if (_dm_bitset)
+		return 1;
+
+	if (!(_dm_bitset = dm_bitset_create(NULL, NUMBER_OF_MAJORS)))
+		return 0;
+
+	if (!_get_proc_number(PROC_DEVICES, DM_NAME, NULL)) {
+		dm_bitset_destroy(_dm_bitset);
+		_dm_bitset = NULL;
+		return 0;
+	}
+
+	return 1;
+#else
+	return 0;
+#endif
+}
+
+int dm_is_dm_major(uint32_t major)
+{
+	if (!_create_dm_bitset())
+		return 0;
+
+	return dm_bit(_dm_bitset, major) ? 1 : 0;
+}
+
 static int _open_control(void)
 {
 #ifdef DM_IOCTLS
@@ -253,6 +298,11 @@ static int _open_control(void)
 		goto error;
 	}
 
+	if (!_create_dm_bitset()) {
+		log_error("Failed to set up list of device-mapper major numbers");
+		return 0;
+	}
+
 	return 1;
 
 error:
@@ -269,27 +319,27 @@ void dm_task_destroy(struct dm_task *dmt)
 
 	for (t = dmt->head; t; t = n) {
 		n = t->next;
-		free(t->params);
-		free(t->type);
-		free(t);
+		dm_free(t->params);
+		dm_free(t->type);
+		dm_free(t);
 	}
 
 	if (dmt->dev_name)
-		free(dmt->dev_name);
+		dm_free(dmt->dev_name);
 
 	if (dmt->newname)
-		free(dmt->newname);
+		dm_free(dmt->newname);
 
 	if (dmt->message)
-		free(dmt->message);
+		dm_free(dmt->message);
 
 	if (dmt->dmi.v4)
-		free(dmt->dmi.v4);
+		dm_free(dmt->dmi.v4);
 
 	if (dmt->uuid)
-		free(dmt->uuid);
+		dm_free(dmt->uuid);
 
-	free(dmt);
+	dm_free(dmt);
 }
 
 /*
@@ -470,7 +520,7 @@ static struct dm_ioctl_v1 *_flatten_v1(struct dm_task *dmt)
 	if (len < min_size)
 		len = min_size;
 
-	if (!(dmi = malloc(len)))
+	if (!(dmi = dm_malloc(len)))
 		return NULL;
 
 	memset(dmi, 0, len);
@@ -519,7 +569,7 @@ static struct dm_ioctl_v1 *_flatten_v1(struct dm_task *dmt)
 	return dmi;
 
       bad:
-	free(dmi);
+	dm_free(dmi);
 	return NULL;
 }
 
@@ -592,7 +642,7 @@ static int _dm_task_run_v1(struct dm_task *dmt)
 
 	dmi = _flatten_v1(dmt);
 	if (!dmi) {
-		log_error("Couldn't create ioctl argument");
+		log_error("Couldn't create ioctl argument.");
 		return 0;
 	}
 
@@ -670,7 +720,7 @@ static int _dm_task_run_v1(struct dm_task *dmt)
 		dmt->type = DM_DEVICE_INFO;
 		if (!dm_task_run(dmt))
 			goto bad;
-		free(dmi);	/* We'll use what info returned */
+		dm_free(dmi);	/* We'll use what info returned */
 		return 1;
 	}
 
@@ -678,7 +728,7 @@ static int _dm_task_run_v1(struct dm_task *dmt)
 	return 1;
 
       bad:
-	free(dmi);
+	dm_free(dmi);
 	return 0;
 }
 
@@ -918,7 +968,7 @@ int dm_task_set_ro(struct dm_task *dmt)
 
 int dm_task_set_newname(struct dm_task *dmt, const char *newname)
 {
-	if (!(dmt->newname = strdup(newname))) {
+	if (!(dmt->newname = dm_strdup(newname))) {
 		log_error("dm_task_set_newname: strdup(%s) failed", newname);
 		return 0;
 	}
@@ -928,7 +978,7 @@ int dm_task_set_newname(struct dm_task *dmt, const char *newname)
 
 int dm_task_set_message(struct dm_task *dmt, const char *message)
 {
-	if (!(dmt->message = strdup(message))) {
+	if (!(dmt->message = dm_strdup(message))) {
 		log_error("dm_task_set_message: strdup(%s) failed", message);
 		return 0;
 	}
@@ -950,6 +1000,13 @@ int dm_task_no_open_count(struct dm_task *dmt)
 	return 1;
 }
 
+int dm_task_skip_lockfs(struct dm_task *dmt)
+{
+	dmt->skip_lockfs = 1;
+
+	return 1;
+}
+
 int dm_task_set_event_nr(struct dm_task *dmt, uint32_t event_nr)
 {
 	dmt->event_nr = event_nr;
@@ -960,7 +1017,7 @@ int dm_task_set_event_nr(struct dm_task *dmt, uint32_t event_nr)
 struct target *create_target(uint64_t start, uint64_t len, const char *type,
 			     const char *params)
 {
-	struct target *t = malloc(sizeof(*t));
+	struct target *t = dm_malloc(sizeof(*t));
 
 	if (!t) {
 		log_error("create_target: malloc(%d) failed", sizeof(*t));
@@ -969,12 +1026,12 @@ struct target *create_target(uint64_t start, uint64_t len, const char *type,
 
 	memset(t, 0, sizeof(*t));
 
-	if (!(t->params = strdup(params))) {
+	if (!(t->params = dm_strdup(params))) {
 		log_error("create_target: strdup(params) failed");
 		goto bad;
 	}
 
-	if (!(t->type = strdup(type))) {
+	if (!(t->type = dm_strdup(type))) {
 		log_error("create_target: strdup(type) failed");
 		goto bad;
 	}
@@ -984,9 +1041,9 @@ struct target *create_target(uint64_t start, uint64_t len, const char *type,
 	return t;
 
       bad:
-	free(t->params);
-	free(t->type);
-	free(t);
+	dm_free(t->params);
+	dm_free(t->type);
+	dm_free(t);
 	return NULL;
 }
 
@@ -1084,7 +1141,7 @@ static struct dm_ioctl *_flatten(struct dm_task *dmt, unsigned repeat_count)
 	while (repeat_count--)
 		len *= 2;
 
-	if (!(dmi = malloc(len)))
+	if (!(dmi = dm_malloc(len)))
 		return NULL;
 
 	memset(dmi, 0, len);
@@ -1105,11 +1162,13 @@ static struct dm_ioctl *_flatten(struct dm_task *dmt, unsigned repeat_count)
 		dmi->flags |= DM_SUSPEND_FLAG;
 	if (dmt->read_only)
 		dmi->flags |= DM_READONLY_FLAG;
+	if (dmt->skip_lockfs)
+		dmi->flags |= DM_SKIP_LOCKFS_FLAG;
 
 	if (dmt->minor >= 0) {
 		if (dmt->major <= 0) {
-			log_error("Missing major number for persistent device");
-			return NULL;
+			log_error("Missing major number for persistent device.");
+			goto bad;
 		}
 		dmi->flags |= DM_PERSISTENT_DEV_FLAG;
 		dmi->dev = MKDEV(dmt->major, dmt->minor);
@@ -1140,7 +1199,7 @@ static struct dm_ioctl *_flatten(struct dm_task *dmt, unsigned repeat_count)
 	return dmi;
 
       bad:
-	free(dmi);
+	dm_free(dmi);
 	return NULL;
 }
 
@@ -1275,7 +1334,7 @@ static int _create_and_load_v4(struct dm_task *dmt)
 
 	/* Use the original structure last so the info will be correct */
 	dmt->type = DM_DEVICE_RESUME;
-	free(dmt->uuid);
+	dm_free(dmt->uuid);
 	dmt->uuid = NULL;
 
 	r = dm_task_run(dmt);
@@ -1290,7 +1349,7 @@ static struct dm_ioctl *_do_dm_ioctl(struct dm_task *dmt, unsigned command,
 
 	dmi = _flatten(dmt, repeat_count);
 	if (!dmi) {
-		log_error("Couldn't create ioctl argument");
+		log_error("Couldn't create ioctl argument.");
 		return NULL;
 	}
 
@@ -1302,17 +1361,25 @@ static struct dm_ioctl *_do_dm_ioctl(struct dm_task *dmt, unsigned command,
 	if (dmt->no_open_count)
 		dmi->flags |= DM_SKIP_BDGET_FLAG;
 
-	log_debug("dm %s %s %s%s%s %c %.0llu %s [%u]",
+	log_debug("dm %s %s %s%s%s %s%0.0d%s%0.0d%s"
+		  "%s%c %.0llu %s [%u]",
 		  _cmd_data_v4[dmt->type].name,
 		  dmi->name, dmi->uuid, dmt->newname ? " " : "",
 		  dmt->newname ? dmt->newname : "",
+		  dmt->major > 0 ? "(" : "",
+		  dmt->major > 0 ? dmt->major : 0,
+		  dmt->major > 0 ? ":" : "",
+		  dmt->minor > 0 ? dmt->minor : 0,
+		  dmt->major > 0 && dmt->minor == 0 ? "0" : "",
+		  dmt->major > 0 ? ") " : "",
 		  dmt->no_open_count ? 'N' : 'O',
 		  dmt->sector, dmt->message ? dmt->message : "",
 		  dmi->data_size);
 #ifdef DM_IOCTLS
 	if (ioctl(_control_fd, command, dmi) < 0) {
 		if (errno == ENXIO && ((dmt->type == DM_DEVICE_INFO) ||
-				       (dmt->type == DM_DEVICE_MKNODES)))
+				       (dmt->type == DM_DEVICE_MKNODES) ||
+				       (dmt->type == DM_DEVICE_STATUS)))
 			dmi->flags &= ~DM_EXISTS_FLAG;	/* FIXME */
 		else {
 			if (_log_suppress)
@@ -1323,7 +1390,7 @@ static struct dm_ioctl *_do_dm_ioctl(struct dm_task *dmt, unsigned command,
 				log_error("device-mapper ioctl "
 					  "cmd %d failed: %s",
 					  _IOC_NR(command), strerror(errno));
-			free(dmi);
+			dm_free(dmi);
 			return NULL;
 		}
 	}
@@ -1375,7 +1442,7 @@ repeat_ioctl:
 		case DM_DEVICE_TABLE:
 		case DM_DEVICE_WAITEVENT:
 			_ioctl_buffer_double_factor++;
-			free(dmi);
+			dm_free(dmi);
 			goto repeat_ioctl;
 		default:
 			log_error("Warning: libdevmapper buffer too small for data");
@@ -1421,7 +1488,7 @@ repeat_ioctl:
 	return 1;
 
       bad:
-	free(dmi);
+	dm_free(dmi);
 	return 0;
 }
 
@@ -1437,6 +1504,10 @@ void dm_lib_release(void)
 void dm_lib_exit(void)
 {
 	dm_lib_release();
+	if (_dm_bitset)
+		dm_bitset_destroy(_dm_bitset);
+	_dm_bitset = NULL;
+	dm_dump_memory();
 	_version_ok = 1;
 	_version_checked = 0;
 }
