@@ -93,7 +93,7 @@ int dm_get_library_version(char *version, size_t size)
 
 struct dm_task *dm_task_create(int type)
 {
-	struct dm_task *dmt = malloc(sizeof(*dmt));
+	struct dm_task *dmt = dm_malloc(sizeof(*dmt));
 
 	if (!dmt) {
 		log_error("dm_task_create: malloc(%d) failed", sizeof(*dmt));
@@ -123,7 +123,7 @@ int dm_task_set_name(struct dm_task *dmt, const char *name)
 	struct stat st1, st2;
 
 	if (dmt->dev_name) {
-		free(dmt->dev_name);
+		dm_free(dmt->dev_name);
 		dmt->dev_name = NULL;
 	}
 
@@ -143,7 +143,7 @@ int dm_task_set_name(struct dm_task *dmt, const char *name)
 		name = pos + 1;
 	}
 
-	if (!(dmt->dev_name = strdup(name))) {
+	if (!(dmt->dev_name = dm_strdup(name))) {
 		log_error("dm_task_set_name: strdup(%s) failed", name);
 		return 0;
 	}
@@ -154,11 +154,11 @@ int dm_task_set_name(struct dm_task *dmt, const char *name)
 int dm_task_set_uuid(struct dm_task *dmt, const char *uuid)
 {
 	if (dmt->uuid) {
-		free(dmt->uuid);
+		dm_free(dmt->uuid);
 		dmt->uuid = NULL;
 	}
 
-	if (!(dmt->uuid = strdup(uuid))) {
+	if (!(dmt->uuid = dm_strdup(uuid))) {
 		log_error("dm_task_set_uuid: strdup(%s) failed", uuid);
 		return 0;
 	}
@@ -169,7 +169,6 @@ int dm_task_set_uuid(struct dm_task *dmt, const char *uuid)
 int dm_task_set_major(struct dm_task *dmt, int major)
 {
 	dmt->major = major;
-	log_debug("Setting major: %d", dmt->major);
 
 	return 1;
 }
@@ -177,7 +176,6 @@ int dm_task_set_major(struct dm_task *dmt, int major)
 int dm_task_set_minor(struct dm_task *dmt, int minor)
 {
 	dmt->minor = minor;
-	log_debug("Setting minor: %d", dmt->minor);
 
 	return 1;
 }
@@ -201,7 +199,7 @@ int dm_task_add_target(struct dm_task *dmt, uint64_t start, uint64_t size,
 }
 
 #ifdef HAVE_SELINUX
-int set_selinux_context(const char *path, mode_t mode)
+int dm_set_selinux_context(const char *path, mode_t mode)
 {
 	security_context_t scontext;
 
@@ -268,7 +266,7 @@ static int _add_dev_node(const char *dev_name, uint32_t major, uint32_t minor,
 	}
 
 #ifdef HAVE_SELINUX
-	if (!set_selinux_context(path, S_IFBLK))
+	if (!dm_set_selinux_context(path, S_IFBLK))
 		return 0;
 #endif
 
@@ -381,7 +379,7 @@ static int _stack_node_op(node_op_t type, const char *dev_name, uint32_t major,
 	size_t len = strlen(dev_name) + strlen(old_name) + 2;
 	char *pos;
 
-	if (!(nop = malloc(sizeof(*nop) + len))) {
+	if (!(nop = dm_malloc(sizeof(*nop) + len))) {
 		log_error("Insufficient memory to stack mknod operation");
 		return 0;
 	}
@@ -412,7 +410,7 @@ static void _pop_node_ops(void)
 		_do_node_op(nop->type, nop->dev_name, nop->major, nop->minor,
 			    nop->uid, nop->gid, nop->mode, nop->old_name);
 		list_del(&nop->list);
-		free(nop);
+		dm_free(nop);
 	}
 }
 
@@ -448,3 +446,45 @@ const char *dm_dir(void)
 {
 	return _dm_dir;
 }
+
+int dm_mknodes(const char *name)
+{
+	struct dm_task *dmt;
+	int r = 0;
+
+	if (!(dmt = dm_task_create(DM_DEVICE_MKNODES)))
+		return 0;
+
+	if (name && !dm_task_set_name(dmt, name))
+		goto out;
+
+	if (!dm_task_no_open_count(dmt))
+		goto out;
+
+	r = dm_task_run(dmt);
+
+out:
+	dm_task_destroy(dmt);
+	return r;
+}
+
+int dm_driver_version(char *version, size_t size)
+{
+	struct dm_task *dmt;
+	int r = 0;
+
+	if (!(dmt = dm_task_create(DM_DEVICE_VERSION)))
+		return 0;
+
+	if (!dm_task_run(dmt))
+		log_error("Failed to get driver version");
+
+	if (!dm_task_get_driver_version(dmt, version, size))
+		goto out;
+
+	r = 1;
+
+out:
+	dm_task_destroy(dmt);
+	return r;
+}
